from enum import Enum
import re
from langchain_openai import ChatOpenAI
from unstructured.partition.pdf import partition_pdf
from dotenv import load_dotenv
import os
from langchain_core.prompts import ChatPromptTemplate
from langchain_community.chat_models import ChatOllama
from langchain_core.output_parsers import StrOutputParser

class ModelEnum(str, Enum):
    """Model to use for the conversion"""
    LOCAL = "llama3"
    GPT4O = "gpt-4o"
    NONE = None

class UnstructuredParser:
    load_dotenv()

    # Function to convert element category to markdown format
    def convert_to_markdown(self, elements):
        markdown_content = ""
        element_hierarchy = {}

        for el in elements:
            markdown_content += self.get_markdown_line(el)

        return markdown_content

    def get_markdown_line(self, el):
        element_type = el["type"]
        text = el["text"]
        metadata = el["metadata"]
        parent_id = metadata.get("parent_id", None)
        category_depth = metadata.get("category_depth", 0)
        if "emphasized_text_contents" in metadata:
            print(metadata["emphasized_text_contents"])

        markdown_line = ""

        if element_type == "Title":
            if parent_id:
                markdown_line = (
                    f"## {text}\n\n"  # Adjusted to add sub headers if parent_id exists
                )
            else:
                markdown_line = f"# {text}\n\n"
        elif element_type == "Subtitle":
            markdown_line = f"## {text}\n\n"
        elif element_type == "Header":
            markdown_line = f"{'#' * (category_depth + 1)} {text}\n\n"
        elif element_type == "Footer":
            markdown_line = f"#### {text}\n\n"
        elif element_type == "NarrativeText":
            markdown_line = f"{text}\n\n"
        elif element_type == "ListItem":
            markdown_line = f"- {text}\n"
        elif element_type == "Table":
            markdown_line = el["metadata"]["text_as_html"]
        elif element_type == "PageBreak":
            markdown_line = f"---\n\n"
        elif element_type == "Image":
            markdown_line = f"![Image]({el['metadata'].get('image_path', '')})\n\n"
        elif element_type == "Formula":
            markdown_line = f"$$ {text} $$\n\n"
        elif element_type == "FigureCaption":
            markdown_line = f"**Figure:** {text}\n\n"
        elif element_type == "Address":
            markdown_line = f"**Address:** {text}\n\n"
        elif element_type == "EmailAddress":
            markdown_line = f"**Email:** {text}\n\n"
        elif element_type == "CodeSnippet":
            markdown_line = f"```{el['metadata'].get('language', '')}\n{text}\n```\n\n"
        elif element_type == "PageNumber":
            markdown_line = f"**Page {text}**\n\n"
        else:
            markdown_line = f"{text}\n\n"

        return markdown_line

    def partition_pdf_file(self, path, strategy="fast"):
        return partition_pdf(
            filename=path, infer_table_structure=True, strategy=strategy
        )

    def improve_layout(self, elements, remove_repeated_headers=True, model: ModelEnum = ModelEnum.GPT4O):
        llm = None
        chain = None
        if model != ModelEnum.NONE:
            llm = ChatOpenAI(model="gpt-4o", temperature=0.1) if model == ModelEnum.GPT4O else ChatOllama(model=model.value,  temperature=0.1)

            # Define the prompt
            prompt = ChatPromptTemplate.from_messages([
                (
                    "human",
                    """You are an expert in markdown tables, match this text and this html table to fill a md table. You answer with just the table in pure markdown, nothing else.
                    <TEXT> 
                    {text} 
                    </TEXT>
                    <HTML>
                    {html}
                    </HTML>
                    
                    Note, the previous table (that might be related since appearing just before):
                    <PREVIOUS_TABLE>
                    {previous_table}
                    </PREVIOUS_TABLE>"""
                ),
            ])
            chain = prompt | llm
        
        table_stack = []

        improved_elements = []
        for el in elements:
            if el.category == "Table":
                if el.text not in set(table_stack):
                    if chain:
                        result = chain.invoke({
                            "text": el.text,
                            "html": el.metadata.text_as_html,
                            "previous_table": table_stack[-1] if table_stack else ""

                        })
                        cleaned_result = result.content
                        cleaned_content = re.sub(r'^```.*$\n?', '', str(cleaned_result), flags=re.MULTILINE)
                    else: 
                        cleaned_content = el.text
                        
                    el.metadata.text_as_html = f"[TABLE]\n{cleaned_content}\n[/TABLE]"
                    # add line break to separate tables
                    el.metadata.text_as_html = el.metadata.text_as_html + "\n\n" #type: ignore
                    table_stack.append(el.text)
                    improved_elements.append(el)

            elif el.category not in ["Header", "Footer"]:
                if "page" not in el.text.lower():
                    if (el.text not in set(table_stack) and "page" not in el.text.lower()) or remove_repeated_headers == False:
                        improved_elements.append(el)

                    table_stack.append(el.text.strip())
                    table_stack.append("")

                

        return improved_elements

    def convert(self, path, model: ModelEnum = ModelEnum.GPT4O, strategy="fast"):
        # Partition the PDF
        elements = self.partition_pdf_file(path, strategy=strategy)

        # Improve table elements
        improved_elements = self.improve_layout(elements, model=model)

        elements_dict = [el.to_dict() for el in improved_elements]
        markdown_content = self.convert_to_markdown(elements_dict)
        return markdown_content


# if __name__ == "__main__":
#     parser = UnstructuredParser()
#     response = parser.convert("megaparse/tests/input_tests/MegaFake_report.pdf", model=ModelEnum.NONE)
#     print(response)
#     with open("megaparse/tests/output_tests/cdp.md", "w") as f:
#         f.write(response)
#     print("ok")


from docx.document import Document as DocumentObject
from docx import Document
from docx.table import Table
from docx.text.paragraph import Paragraph
from docx.section import Section, _Header as Header, _Footer as Footer
from docx.oxml.text.paragraph import CT_P
from docx.oxml.table import CT_Tbl

def print_element(element):
    if isinstance(element, Paragraph):
        # Print the paragraph text
        print(f'Paragraph: {element.text}')
    elif isinstance(element, Table):
        # Print the table content
        print('Table:')
        for row in element.rows:
            for cell in row.cells:
                print(cell.text, end='\t')
            print()
    elif isinstance(element, Section):
        # Print section properties
        print('Section:')
        print(f'  Start type: {element.start_type}')
        print(f'  Page height: {element.page_height}')
        print(f'  Page width: {element.page_width}')
    elif isinstance(element, Header):
        # Print header content
        print('Header:')
        for paragraph in element.paragraphs:
            print(f'  {paragraph.text}')
    elif isinstance(element, Footer):
        # Print footer content
        print('Footer:')
        for paragraph in element.paragraphs:
            print(f'  {paragraph.text}')
    else:
        print(f'Unknown element: {type(element)}')


def print_docx(doc: DocumentObject) -> None:
    for element in doc.element.body:
        if isinstance(element, CT_P):  # Paragraph
            print_element(Paragraph(element, doc))
        elif isinstance(element, CT_Tbl):  # Table
            print_element(Table(element, doc))





import asyncio
import os
from collections import Counter
from enum import Enum
from pathlib import Path
from typing import List, Set

import pandas as pd
from docx import Document
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P
from docx.table import Table
from docx.text.paragraph import Paragraph
from docx.text.run import Run
from langchain_community.document_loaders.base import BaseLoader
from langchain_core.documents import Document as LangChainDocument
from llama_index.core.schema import Document as LlamaDocument
from llama_parse import LlamaParse
from llama_parse.utils import Language, ResultType
from pptx import Presentation
from pptx.enum.shapes import MSO_SHAPE_TYPE

from megaparse.markdown_processor import MarkdownProcessor
from megaparse.multimodal_convertor.megaparse_vision import MegaParseVision
from megaparse.unstructured_convertor import ModelEnum, UnstructuredParser


class Converter:
    def __init__(self) -> None:
        pass

    async def convert(self, file_path: str | Path) -> LangChainDocument:
        raise NotImplementedError("Subclasses should implement this method")

    def save_md(self, md_content: str, file_path: Path | str) -> None:
        with open(file_path, "w") as f:
            f.write(md_content)


class XLSXConverter(Converter):
    def __init__(self) -> None:
        pass

    async def convert(self, file_path: str | Path) -> LangChainDocument:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        xls = pd.ExcelFile(file_path)  # type: ignore
        sheets = pd.read_excel(xls)

        target_text = self.table_to_text(sheets)

        return LangChainDocument(
            page_content=target_text,
            metadata={"filename": file_path.name, "type": "xlsx"},
        )

    def convert_tab(self, file_path: str | Path, tab_name: str) -> str:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        xls = pd.ExcelFile(str(file_path))
        sheets = pd.read_excel(xls, tab_name)
        target_text = self.table_to_text(sheets)
        return target_text

    def table_to_text(self, df):
        text_rows = []
        for _, row in df.iterrows():
            row_text = " | ".join(str(value) for value in row.values if pd.notna(value))
            if row_text:
                text_rows.append("|" + row_text + "|")
        return "\n".join(text_rows)


class DOCXConverter(Converter):
    def __init__(self) -> None:
        self.header_handled = False

    async def convert(self, file_path: str | Path) -> LangChainDocument:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        doc = Document(str(file_path))
        md_content = []
        # Handle header
        if doc.sections and doc.sections[0].header:
            header_content = self._handle_header(doc.sections[0].header)
            if header_content:
                md_content.append(header_content)

        for element in doc.element.body:
            if isinstance(element, CT_P):
                md_content.append(self._handle_paragraph(Paragraph(element, doc)))
            elif isinstance(element, CT_Tbl):
                md_content += self._handle_table(Table(element, doc))
            # Add more handlers here (image, header, footer, etc)

        return LangChainDocument(
            page_content="\n".join(md_content),
            metadata={"filename": file_path.name, "type": "docx"},
        )

    def _handle_header(self, header) -> str:
        if not self.header_handled:
            parts = []
            for paragraph in header.paragraphs:
                parts.append(f"# {paragraph.text}")
            for table in header.tables:
                parts += self._handle_header_table(table)
            self.header_handled = True
            return "\n".join(parts)
        return ""

    def _handle_header_table(self, table: Table) -> List[str]:
        cell_texts = [cell.text for row in table.rows for cell in row.cells]
        cell_texts.remove("")
        # Find the most repeated cell text
        text_counts = Counter(cell_texts)
        title = text_counts.most_common(1)[0][0] if cell_texts else ""
        other_texts = [text for text in cell_texts if text != title and text != ""]

        md_table_content = []
        if title:
            md_table_content.append(f"# {title}")
        for text in other_texts:
            md_table_content.append(f"*{text}*;")
        return md_table_content

    def _handle_paragraph(self, paragraph: Paragraph) -> str:
        if paragraph.style.name.startswith("Heading"):  # type: ignore
            level = int(paragraph.style.name.split()[-1])  # type: ignore
            return f"{'#' * level} {paragraph.text}"
        else:
            parts = []
            for run in paragraph.runs:
                if run.text != "":
                    parts.append(self._handle_run(run))
            return "".join(parts)

    def _handle_run(self, run: Run) -> str:
        text: str = run.text
        if run.bold:
            if len(text) < 200:
                # FIXME : handle table needs to be improved -> have the paragraph they are in
                text = f"## {text}"
            else:
                text = f"**{text}**"
        if run.italic:
            text = f"*{text}*"
        return text

    def _handle_table(self, table: Table) -> List[str]:
        row_content = []
        for i, row in enumerate(table.rows):
            row_content.append(
                "| " + " | ".join(cell.text.strip() for cell in row.cells) + " |"
            )
            if i == 0:
                row_content.append("|" + "---|" * len(row.cells))

        return row_content

    def save_md(self, md_content: str, file_path: Path | str) -> None:
        with open(file_path, "w") as f:
            f.write(md_content)


class PPTXConverter:
    def __init__(self, add_images=False) -> None:
        self.header_handled = False
        self.add_images = add_images

    async def convert(self, file_path: str | Path) -> LangChainDocument:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        prs = Presentation(str(file_path))
        md_content = []
        unique_slides: Set[str] = set()

        # Handle header
        if prs.slides and prs.slides[0].placeholders:
            header_content = self._handle_header(prs.slides[0].placeholders)
            if header_content:
                md_content.append(header_content)

        for i, slide in enumerate(prs.slides):
            slide_md_content: List[str] = []
            for shape in slide.shapes:
                if shape.shape_type == MSO_SHAPE_TYPE.TABLE:  # type: ignore
                    slide_md_content += self._handle_table(shape.table)
                elif shape.shape_type == MSO_SHAPE_TYPE.PICTURE and self.add_images:  # type: ignore
                    slide_md_content.append(self._handle_image(shape))
                elif hasattr(shape, "text"):
                    slide_md_content.append(self._handle_paragraph(shape.text))

            slide_md_str = "\n".join(slide_md_content)
            if slide_md_str not in unique_slides:
                unique_slides.add(slide_md_str)
                slide_md_str = f"## Slide {i+1}\n{slide_md_str}"
                md_content.append(slide_md_str)

        return LangChainDocument(
            page_content="\n".join(md_content),
            metadata={"filename": file_path.name, "type": "pptx"},
        )

    def _handle_header(self, placeholders) -> str:
        if not self.header_handled:
            parts = []
            for placeholder in placeholders:
                if placeholder.placeholder_format.idx == 0:  # Title placeholder
                    parts.append(f"# {placeholder.text}")
                elif placeholder.placeholder_format.idx == 1:  # Subtitle placeholder
                    parts.append(f"## {placeholder.text}")
            self.header_handled = True
            return "\n".join(parts)
        return ""

    def _handle_paragraph(self, text: str) -> str:
        # Assuming text is a simple paragraph without complex formatting
        # if text contains letters return text
        if any(c.isalpha() for c in text):
            return text + "\n"
        return ""

    def _handle_image(self, shape) -> str:
        image = shape.image
        image_bytes = image.blob
        image_format = image.ext
        image_filename = f"images/image_{shape.shape_id}.{image_format}"
        with open(image_filename, "wb") as f:
            f.write(image_bytes)
        return f"![Image {shape.shape_id}](../{image_filename})"

    def _handle_table(self, table) -> List[str]:
        row_content = []
        for i, row in enumerate(table.rows):
            row_content.append(
                "| " + " | ".join(cell.text.strip() for cell in row.cells) + " |"
            )
            if i == 0:
                row_content.append("|" + "---|" * len(row.cells))
        return row_content

    def save_md(self, md_content: str, file_path: Path | str) -> None:
        with open(file_path, "w") as f:
            f.write(md_content)


class MethodEnum(str, Enum):
    """Method to use for the conversion"""

    LLAMA_PARSE = "llama_parse"
    UNSTRUCTURED = "unstructured"
    MEGAPARSE_VISION = "megaparse_vision"


class PDFConverter:
    def __init__(
        self,
        llama_parse_api_key: str,
        method: MethodEnum | str = MethodEnum.UNSTRUCTURED,
        model=ModelEnum.NONE,
        strategy="fast",
    ) -> None:
        self.strategy = strategy
        self.llama_parse_api_key = llama_parse_api_key
        if isinstance(method, str):
            try:
                method = MethodEnum(method)
            except ValueError:
                raise ValueError(f"Method {method} not supported")
        self.method = method

    async def _llama_parse(self, api_key: str, file_path: str | Path):
        parsing_instructions = "Do not take into account the page breaks (no --- between pages), do not repeat the header and the footer so the tables are merged. Keep the same format for similar tables."
        self.parser = LlamaParse(
            api_key=str(api_key),
            result_type=ResultType.MD,
            gpt4o_mode=True,
            verbose=True,
            language=Language.FRENCH,
            parsing_instruction=parsing_instructions,  # Optionally you can define a parsing instruction
        )
        documents: List[LlamaDocument] = await self.parser.aload_data(str(file_path))
        parsed_md = ""
        for document in documents:
            text_content = document.text
            parsed_md = parsed_md + text_content
        return parsed_md

    def _unstructured_parse(
        self, file_path: str | Path, model: ModelEnum = ModelEnum.NONE
    ):
        unstructured_parser = UnstructuredParser()
        return unstructured_parser.convert(
            file_path, model=model, strategy=self.strategy
        )

    async def _lmm_parse(self, file_path: str | Path):
        lmm_parser = MegaParseVision()
        return await lmm_parser.parse(file_path)

    async def convert(
        self,
        file_path: str | Path,
        model: ModelEnum = ModelEnum.NONE,
        gpt4o_cleaner=False,
    ) -> LangChainDocument:
        if isinstance(file_path, str):
            file_path = Path(file_path)

        parsed_md = ""
        if self.method == MethodEnum.LLAMA_PARSE:
            assert (
                self.llama_parse_api_key is not None
            ), "LLama Parse API key is required for this method"
            parsed_md = await self._llama_parse(self.llama_parse_api_key, file_path)
        elif self.method == MethodEnum.MEGAPARSE_VISION:
            parsed_md = await self._lmm_parse(file_path)
        elif self.method == MethodEnum.UNSTRUCTURED:
            parsed_md = self._unstructured_parse(file_path, model)
        else:
            raise ValueError(f"Method {self.method} not supported")

        if not gpt4o_cleaner:
            return LangChainDocument(
                page_content=parsed_md,
                metadata={"filename": file_path.name, "type": "pdf"},
            )
        else:
            md_processor = MarkdownProcessor(
                parsed_md,
                strict=True,
                remove_pagination=True,
            )
            md_cleaned = md_processor.process(gpt4o_cleaner=gpt4o_cleaner)
            return LangChainDocument(
                page_content=md_cleaned,
                metadata={"filename": file_path.name, "type": "pdf"},
            )

    def save_md(self, md_content: str, file_path: Path | str) -> None:
        with open(file_path, "w") as f:
            f.write(md_content)


class MegaParse(BaseLoader):
    def __init__(
        self,
        file_path: str | Path,
        llama_parse_api_key: str | None = None,
        strategy="fast",
    ) -> None:
        if isinstance(file_path, str):
            file_path = Path(file_path)
        self.file_path = file_path
        self.llama_parse_api_key = llama_parse_api_key
        self.strategy = strategy

    async def aload(self, **convert_kwargs) -> LangChainDocument:
        file_extension: str = os.path.splitext(self.file_path)[1]
        if file_extension == ".docx":
            converter = DOCXConverter()
        elif file_extension == ".pptx":
            converter = PPTXConverter()
        elif file_extension == ".pdf":
            converter = PDFConverter(
                llama_parse_api_key=str(self.llama_parse_api_key),
                strategy=self.strategy,
            )
        elif file_extension == ".xlsx":
            converter = XLSXConverter()
        else:
            raise ValueError(f"Unsupported file extension: {file_extension}")

        return await converter.convert(self.file_path, **convert_kwargs)

    def load(self, **kwargs) -> LangChainDocument:
        file_extension: str = os.path.splitext(self.file_path)[1]
        if file_extension == ".docx":
            converter = DOCXConverter()
        elif file_extension == ".pptx":
            converter = PPTXConverter()
        elif file_extension == ".pdf":
            converter = PDFConverter(
                llama_parse_api_key=str(self.llama_parse_api_key),
                strategy=self.strategy,
            )
        elif file_extension == ".xlsx":
            converter = XLSXConverter()
        else:
            print(self.file_path, file_extension)
            raise ValueError(f"Unsupported file extension: {file_extension}")

        loop = asyncio.get_event_loop()
        return loop.run_until_complete(converter.convert(self.file_path, **kwargs))

    def load_tab(self, tab_name: str, **kwargs) -> LangChainDocument:
        file_extension: str = os.path.splitext(self.file_path)[1]
        if file_extension == ".xlsx":
            converter = XLSXConverter()
        else:
            print(self.file_path, file_extension)
            raise ValueError(f"Unsupported file extension for tabs: {file_extension}")

        result = converter.convert_tab(self.file_path, tab_name=tab_name)
        return LangChainDocument(
            page_content=result,
            metadata={"filename": self.file_path.name, "type": "xlsx"},
        )

    def save_md(self, md_content: str, file_path: Path | str) -> None:
        os.makedirs(os.path.dirname(file_path), exist_ok=True)
        with open(file_path, "w+") as f:
            f.write(md_content)
